﻿// Generator DDS z wobulatorem - 3rd ed.
// AVT-5980 - by Adam Sobczyk SQ5RWQ
// Ver. 1.0, 2023-02-16


// Fusbit-y do zaprogramowania w MCU AVR ATmega328: Lock:FF, Ext:FF, High:D7, Low:F7
// Komenda: sudo avrdude -p m328p -c usbasp -U lfuse:w:0xF7:m -U hfuse:w:0xD7:m -U efuse:w:0xFF:m -U lock:w:0xFF:m -F -P usb
// https://eleccelerator.com/fusecalc/fusecalc.php?chip=atmega328p&LOW=F7&HIGH=D7&EXTENDED=FF&LOCKBIT=FF
// Częstotliwość F_CPU (=16MHz): zdefiniowana takze w pliku 'Makefile'
// Zaprogramowanie MCU
// Komenda: sudo avrdude -p m328p -c usbasp -U flash:w:mainfile.hex:i -F -P usb


/****************************** Makrodefinicje stalych symbolicznych ******************************/
#define F_CPU 16000000UL	// 16 MHz
#define Fg_tab_MAX 256		// liczba probek Fg dla wobulatora


/********************************* Deklaracje plikow naglowkowych *********************************/
#include <stdio.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include "AVT-5980-lcd.h"
#include "AVT-5980-functions.h"


/************************** Deklaracje parametrow i zmiennych globalnych **************************/
char linebuf[17] = ""; 							// bufor znakowy dla wyświetlacza LCD 16x2 (jedna linia)
const char welcome[4][17] =					// linie tekstu powitalnego
{
	" Stepped sweep ", " DDS generator ",
	"   AVT-5980    ", "by Adam SQ5RWQ "
};
char enc_new, enc_old;   						// biezacy (nowo odczytany) oraz poprzedni stan enkodera
char enc_state;              				// kombinacja  enc_new (b1b0) oraz enc_old (b3b2) w jednej zmiennej
char enc_todo=0x00;      						// wynik analizy robionej w przerwaniu: co zadzialo sie z enkoderem: 1 -> w prawo, 2 -> w lewo
uint8_t eeprom_stamp;								// znacznik stanu zapisu startowych parametrow konfiguracyjnych (0x00 przy pierwszym starcie MCU)
uint8_t mode; 											// tryb pracy: 0 - generator; 1 - wobulator
uint32_t Fg;												// aktualna F generatora [Hz]
uint32_t Fg_old;										// poprzednia F generatora [Hz]
uint32_t dFg;            						// krok regulacji czestotliwosci [Hz]
int dFg_ind;												// index do tablicy z wartosciami dFg (wartosci: 0..18)
const uint32_t dFg_tab[]={1, 2, 5, 10, 20, 50, 100, 200, 500,
			1000, 2000, 5000, 10000, 20000, 50000,
			100000, 200000, 500000, 1000000};	// dopuszczalne wartosci dFg
uint32_t Fl;												// aktualna F dolna wobulatora [Hz]
uint32_t Fl_old;										// poprzednia F dolna wobulatora [Hz]
uint32_t Fh;												// aktualna F gorna wobulatora [Hz]
uint32_t Fh_old;										// poprzednia F gorna wobulatora [Hz]
uint16_t Fw;												// czestotliwosc wobulacji [Hz]
uint16_t Tw;												// elementarne opoznienie wobulacji [???]
int Fw_ind;													// index do tablic z wartosciami Fw i Tw  (wartosci: 0..6)
const uint16_t Fw_tab[]={4, 6, 8, 10};				// dopuszczalne wartosci Fw [Hz]
const uint16_t Tw_tab[]={170, 88, 48, 23};		// dopuszczalne wartosci Tw (krotnosci opoznienia jednostkowego, rownego ok. 4us)
uint32_t Fg_tab[Fg_tab_MAX];				// tablica na probki F dla wobulatora
float Vout;													// napiecie zmierzone na wyjsciu analogowym


/*************************************** Glowny blok programu *****************************************/
int main(void)
{
	/************************************* Konfiguracje portow ************************************/
	// PORTD: linie DB.4..DB.7 wyświetlacza LCD 2x16  -> PD.4..PD.7 (OUT)
	// PORTD: linie RS, R/W i E wyświetlacza LCD 2x16 -> PD.0..PD.2 (OUT)
	// PORTD: linia INT (IN) -> PD.3 (IN)
	DDRD  = 0b11110111; PORTD = 0b00000000; 	// DDRD = 0xF7; PORTD = 0x00;

	// PORTB: linie PB.0,PB.1->ENCA,ENCB (IN); PB.2->DAC/PWM/OC1B (tymczasowo IN); PB.3..PB.5->MOSI/MISO/SCK (n/u)
	DDRB =  0b00000000; PORTB = 0b00000000;		// DDRB = 0x00; PORTB = 0x00;

	// PORTC: PC.1->ENCS (IN/pull-up); PC.2->DDS_RESET (OUT); PC.3->DDS_DATA (OUT); PC.4->DDS_FQ_UD (OUT); PC.5->DDS_W_CLK (OUT);
	DDRC =  0b00111100; PORTC = 0b00000010;		// DDRC = 0x3C; PORTC = 0x02;


	/******************** Inicjalizacja pracy ekranu LCD, enkodera i modulu DDS ********************/
	InitLcd(); 											// inicjalizacja LCD
	enc_new = PORTB & 0b00000011;   // poczatkowa wartosc zmiennej enc_new (istotne tylko PB.1 i PB.0)
	ext_int1_init();                // inicjalizacja zewnetrznego przerwania INT1
	sei();													// globalna zgoda na przerwania
	dds_init();											// inicjalizacja modulu DDS
	dds_freq_set(0);								// zablokowanie pracy modulu DDS


	/******************** Ustawianie poczatkowych wartosci kluczowych parametrow ********************/
	eeprom_stamp = eeprom_read_byte((uint8_t*) 0x01);		// odczyt znacznikow zapisu parametrow konfiguracyjnych

	if(eeprom_stamp & 0b00000001)												// test - czy byl juz zapisywany parametr mode ?
		mode = eeprom_read_byte((uint8_t*) 0x02);
	else
		mode = 0;								// default na 1-y start MCU
	
	if(eeprom_stamp & 0b00000010)												// test - czy byla juz zapisywana Fg ?
		Fg = eeprom_read_dword((uint32_t*) 0x03);
	else
		Fg = 1000;							// default na 1-y start MCU
	Fg_old = 100000000;				// poprzednia F generatora [Hz] - ustawienie poczatkowe "dummy"

	if(eeprom_stamp & 0b00000100)												// test - czy byl juz zapisywany dFg_ind ?
		dFg_ind = eeprom_read_word((uint16_t*) 0x07);
	else
		dFg_ind = 2;						// default na 1-y start MCU
	dFg = dFg_tab[dFg_ind]; 	// ustawienie wartosci dFg

	if(eeprom_stamp & 0b00001000)												// test - czy byla juz zapisywana Fl ?
		Fl = eeprom_read_dword((uint32_t*) 0x09);
	else
		Fl = 20;								// default na 1-y start MCU
	Fl_old = 100000000;				// poprzednia Fl wobulatora - ustawienie poczatkowe "dummy"

	if(eeprom_stamp & 0b00010000)												// test - czy byla juz zapisywana Fh ?
		Fh = eeprom_read_dword((uint32_t*) 0x0D);
	else
		Fh = 20000;							// default na 1-y start MCU
	Fh_old = 100000000;				// poprzednia Fh wobulatora - ustawienie poczatkowe "dummy"

	if(eeprom_stamp & 0b00100000)												// test - czy byl juz zapisywany Fw_ind ?
		Fw_ind = eeprom_read_word((uint16_t*) 0x11);
	else
		Fw_ind = 0;							// default na 1-y start MCU
	Fw = Fw_tab[Fw_ind];  		// czestotliwosc wobulacji - ustawienie poczatkowe
	Tw = Tw_tab[Fw_ind];  		// elementarne opoznienie wobulacji - ustawienie poczatkowe


	/******************************* Ekrany powitalno-informacyjne **********************************/
	snprintf(linebuf, sizeof(linebuf), welcome[0]);
	InstLcd(LCDLINE1); PutsLcd(linebuf);									// LCD out
	snprintf(linebuf, sizeof(linebuf), welcome[1]);
	InstLcd(LCDLINE2); PutsLcd(linebuf);									// LCD out
	while(PINC & _BV(PC1)); while((PINC & _BV(PC1))==0);	// czekanie na nacisniecie przycisku oraz test jego zwolnienia

	snprintf(linebuf, sizeof(linebuf), welcome[2]);
	InstLcd(LCDLINE1); PutsLcd(linebuf);									// LCD out
	snprintf(linebuf, sizeof(linebuf), welcome[3]);
	InstLcd(LCDLINE2); PutsLcd(linebuf);									// LCD out
	while(PINC & _BV(PC1)); while((PINC & _BV(PC1))==0);	// czekanie na nacisniecie przycisku oraz test jego zwolnienia
		

	/************************************** Wybor trybu pracy ***************************************/
	snprintf(linebuf, sizeof(linebuf), "Tryb pracy:     ");
	InstLcd(LCDLINE1); PutsLcd(linebuf);	// LCD out
	while(PINC & _BV(PC1))								// nacisniecie przycisku przerywa petle
	{
	        switch(enc_todo)
	        {
	        	case 1 : mode=0; break;			// wybrano pracę z generatorem
	        	case 2 : mode=1; break;			// wybrano pracę z wobulatorem
	        }
	        enc_todo=0;										// zgaszenie flagi "wyjscia" enkodera
	        
		switch(mode)
		{
			case 0 : snprintf(linebuf, sizeof(linebuf), "    GENERATOR   "); break;
			case 1 : snprintf(linebuf, sizeof(linebuf), "    WOBULATOR   "); break;
		}
		InstLcd(LCDLINE2); PutsLcd(linebuf);							// LCD out
	}
	while((PINC & _BV(PC1))==0);												// test zwolnienia przycisku
	eeprom_update_byte((uint8_t*) 0x02, mode);					// aktualizacja parametru mode w EEPROM
	eeprom_stamp|=0b00000001;														// modyfikacja eeprom_stamp
	eeprom_update_byte((uint8_t*) 0x01, eeprom_stamp);	// aktualizacja eeprom_stamp w EEPROM


	/*********************** Rozgalezienie kodu na generator albo wobulator ************************/
	if(mode==0)			// tryb pracy generatora
	{
		/*********************************** Obsluga generatora ************************************/
		dds_init(); 				// inicjalizacja DDS-a
		dds_freq_set(Fg);		// wstepne ustawienie Fg w module DDS
		adc_init();					// inicjalizacja ADC
		
		while(1)
		{
			/********** Normalna praca: pomiar Vout i zmiana Fg **********/
			F_show(1, Fg);						// wyswietlanie aktualnej Fg
			while(PINC & _BV(PC1))		// nacisniecie przycisku przerywa petle
			{
				Vout=Vout_measure(); Vout_show(Vout);														// pomiar i wyswietlenie aktualnego Vout
				if(enc_todo==1) {Fg_old=Fg; if((Fg+dFg)<=40000000UL) Fg+=dFg;}	// warunkowe zwiekszenie Fg (z ograniczeniem)
				if(enc_todo==2) {Fg_old=Fg; if(Fg>dFg) Fg-=dFg;}								// warunkowe zmniejszenie Fg (z ograniczeniem)
				enc_todo=0;																											// zgaszenie flagi "wyjscia" enkodera
				if(Fg!=Fg_old) {dds_freq_set(Fg); F_show(1, Fg);}								// zmiana Fg -> ustawienie w module DDS i wyswietlenie nowej Fg
			}
			while((PINC & _BV(PC1))==0);												// test zwolnienia przycisku
			eeprom_update_dword((uint32_t*) 0x03, Fg);					// aktualizacja Fg w EEPROM
			eeprom_stamp|=0b00000010;														// modyfikacja eeprom_stamp
			eeprom_update_byte((uint8_t*) 0x01, eeprom_stamp);	// aktualizacja eeprom_stamp w EEPROM
			
			/**************** Zmiana kroku regulacji dFg ****************/
			dFg_show(dFg);								// wyswietlenie aktualnego dFg
			while(PINC & _BV(PC1))				// nacisniecie przycisku przerywa petle
			{
				if(enc_todo==1) dFg_ind++;																// zwiekszenie dFg_ind
				if(enc_todo==2) dFg_ind--;																// zmniejszenie dFg_ind
				if(dFg_ind<0) dFg_ind=0; else if(dFg_ind>18) dFg_ind=18;	// korekta dFg_ind do dozwolonego zakresu
				enc_todo=0;																								// zgaszenie flagi "wyjscia" enkodera
				dFg=dFg_tab[dFg_ind]; dFg_show(dFg);											// wyznaczenie i wyswietlenie aktualnego dFg
			}
			while((PINC & _BV(PC1))==0);												// test zwolnienia przycisku
			eeprom_update_word((uint16_t*) 0x07, dFg_ind);			// aktualizacja dFg_ind w EEPROM
			eeprom_stamp|=0b00000100;														// modyfikacja eeprom_stamp
			eeprom_update_byte((uint8_t*) 0x01, eeprom_stamp);	// aktualizacja eeprom_stamp w EEPROM
		}
	}
	else			// tryb pracy wobulatora
	{
		//*********************************** Obsluga wobulatora ************************************
		dds_init(); 		// inicjalizacja modulu DDS
		adc_init();			// inicjalizacja ADC
		pwm_init();			// konfiguracja i inicjalizacja PWM/OCR1B/PB.2
		OCR1B = 0;			// ustawienie 0V na wyjsciu SYNCHRO
		tc0_config();		// konfiguracja timera TC0
		
		while(1)
		{
			/********** Normalna praca: wyswietlenie Fl, Fh i wobulacja **********/
			F_show(2, Fl);		// wyswietlenie aktualnej Fl
			F_show(3, Fh);		// wyswietlenie aktualnej Fh

			for(uint16_t i=0; i<Fg_tab_MAX; i++) Fg_tab[i] = (uint32_t) (Fl + ((Fh-Fl))*i/255); // tablicowanie Fg
			cli();				// globalna blokada przerwan dla unikniecia zaburzenia konfiguracji PWM1B i TC0
			OCR0A = Tw;		// ustawienie po konfiguracji TC0 ale przed jego startem
			for(;;)
			{
				for(uint16_t pwm_val=0; pwm_val<Fg_tab_MAX; pwm_val++)	// wartosci sterujace procesem wobulacji
				{
					OCR1B = pwm_val;									// ustawienie Fast_PWM (DAC)
					dds_freq_set(Fg_tab[pwm_val]);		// ustawienie czestotliwosci pracy modulu DDS
					tc0_start();											// start TC0, OCR0A ustawione po ustawieniu Fw_ind
					while(!(TIFR0 & (1<<OCF0A))) {};	// iterowanie az TCNT0==OCR0A
					tc0_stop();												// stop TC0
					if((PINC & _BV(PC1))==0) break;		// nacisniecie przycisku przerywa petle
				}
				dds_freq_set(0);									// wylaczenie DDS-a
				OCR1B = 0;												// 0V na wyjsciu Fast_PWM (DAC)
				if((PINC & _BV(PC1))==0) break;		// nacisniecie przycisku przerywa petle
			}
			while((PINC & _BV(PC1))==0);	// test zwolnienia przycisku
			sei();												// globalne odblokowanie przerwan

			/**************** Zmiana kroku regulacji dFg ****************/
			dFg_show(dFg);						// wyswietlenie aktualnego dFg
			while(PINC & _BV(PC1))		// nacisniecie przycisku przerywa petle
			{
				if(enc_todo==1) dFg_ind++;																// zwiekszenie dFg_ind
				if(enc_todo==2) dFg_ind--;																// zmniejszenie dFg_ind
				if(dFg_ind<0) dFg_ind=0; else if(dFg_ind>18) dFg_ind=18;	// korekta dFg_ind do dozwolonego zakresu
				enc_todo=0;																								// zgaszenie flagi "wyjscia" enkodera
				dFg=dFg_tab[dFg_ind]; dFg_show(dFg);											// wyznaczenie i wyswietlenie aktualnego dFg
			}
			while((PINC & _BV(PC1))==0);												// test zwolnienia przycisku
			eeprom_update_word((uint16_t*) 0x07, dFg_ind);			// aktualizacja dFg_ind w EEPROM
			eeprom_stamp|=0b00000100;														// modyfikacja eeprom_stamp
			eeprom_update_byte((uint8_t*) 0x01, eeprom_stamp);	// aktualizacja eeprom_stamp w EEPROM
			
			/************************ Zmiana Fl *************************/
			snprintf(linebuf, sizeof(linebuf), "----------------"); // kasowanie dolnej linii LCD
			InstLcd(LCDLINE2); PutsLcd(linebuf);										// LCD out
			F_show(2, Fl);																					// wyswietlenie aktualnej Fl
			while(PINC & _BV(PC1))																	// nacisniecie przycisku przerywa petle
			{
				if(enc_todo==1) {Fl_old=Fl; if((Fl+dFg)<Fh) Fl+=dFg;}							// warunkowe zwiekszenie Fl (z ograniczeniem)
				if(enc_todo==2) {Fl_old=Fl; if(Fl>dFg) Fl-=dFg;}									// warunkowe zmniejszenie Fl (z ograniczeniem)
				enc_todo=0;																												// zgaszenie flagi "wyjscia" enkodera
				if(Fl!=Fl_old) F_show(2, Fl);																			// wyswietlenie zmodyfikowanej Fl
			}
			while((PINC & _BV(PC1))==0);												// test zwolnienia przycisku
			eeprom_update_dword((uint32_t*) 0x09, Fl);					// aktualizacja Fl w EEPROM
			eeprom_stamp|=0b00001000;														// modyfikacja eeprom_stamp
			eeprom_update_byte((uint8_t*) 0x01, eeprom_stamp);	// aktualizacja eeprom_stamp w EEPROM
			
			/************************ Zmiana Fh *************************/
			snprintf(linebuf, sizeof(linebuf), "----------------"); // kasowanie gornej linii LCD
			InstLcd(LCDLINE1); PutsLcd(linebuf);										// LCD out
			F_show(3, Fh);																					// wyswietlanie aktualnej Fh
			while(PINC & _BV(PC1))																	// nacisniecie przycisku przerywa petle
			{
				if(enc_todo==1) {Fh_old=Fh; if((Fh+dFg)<=40000000UL) Fh+=dFg;}		// warunkowe zwiekszenie Fh (z ograniczeniem)
				if(enc_todo==2) {Fh_old=Fh; if(Fh>(Fl+dFg)) Fh-=dFg;}							// warunkowe zmniejszenie Fh (z ograniczeniem)
				enc_todo=0;																												// zgaszenie flagi "wyjscia" enkodera
				if(Fh!=Fh_old) F_show(3, Fh);																			// wyswietlenie zmodyfikowanej Fh 
			}
			while((PINC & _BV(PC1))==0);												// test zwolnienia przycisku
			eeprom_update_dword((uint32_t*) 0x0D, Fh);					// aktualizacja Fh w EEPROM
			eeprom_stamp|=0b00010000;														// modyfikacja eeprom_stamp
			eeprom_update_byte((uint8_t*) 0x01, eeprom_stamp);	// aktualizacja eeprom_stamp w EEPROM

			/********** Zmiana czestotliwosci wobulacji Fw=f(Tw) **********/
			Fw_show(Fw);									// wyswietlenie aktualnej Fw
			while(PINC & _BV(PC1))				// nacisniecie przycisku przerywa petle
			{
				if(enc_todo==1) if(Fw_ind<3) 	Fw_ind++;			// kontrolowane zwiekszenie Fw_ind
				if(enc_todo==2) if(Fw_ind>0) 	Fw_ind--;			// kontrolowane zmniejszenie Fw_ind
				enc_todo=0;																	// zgaszenie flagi "wyjscia" enkodera
				Fw = Fw_tab[Fw_ind];				// wyznaczenie aktualnej Fw
				Tw = Tw_tab[Fw_ind];				// wyznaczenie aktualnego Tw
				Fw_show(Fw);								// wyswietlenie aktualnej Fw
			}
			while((PINC & _BV(PC1))==0);												// test zwolnienia przycisku
			eeprom_update_word((uint16_t*) 0x11, Fw_ind);				// aktualizacja Fw_ind w EEPROM
			eeprom_stamp|=0b00100000;														// modyfikacja eeprom_stamp
			eeprom_update_byte((uint8_t*) 0x01, eeprom_stamp);	// aktualizacja eeprom_stamp w EEPROM

			/**** Regulacja i pomiar Vout na srodku przemiatanego pasma ****/
			Fg=(Fl+Fh)/2; 								// ustawienie Fg na srodku przemiatanego pasma
			dds_freq_set(Fg);							// ustawienie nowej Fg w module DDS
			F_show(1, Fg);								// wyswietlenie aktualnej Fg
			while(PINC & _BV(PC1))				// nacisniecie przycisku przerywa petle
			{
				Vout=Vout_measure();				// pomiar aktualnego Vout
				Vout_show(Vout);						// wyswietlenie aktualnego Vout
			}
			while((PINC & _BV(PC1))==0);	// test zwolnienia przycisku
		}
	}
	return 0; 	// teoretyczne zakonczenie programu (z uwagi na nieskonczone petle nie powinno nigdy nastapic)
}

